
import calendar
import os
import warnings
import scipy

import cartopy.crs as ccrs
import cdsapi
import numpy as np
import pandas as pd
import xarray as xr

import ddeq

from ddeq.smartcarb import DOMAIN


def read_smartcarb(time, lon, lat, radius=None, data_path='.', method='linear',
                   average=False):
    """\
    Read SMARTCARB winds at given location at given `time`. The location is
    interpolated from SMARTCARB model grid to given `lon` and `lat`. If `lon`
    and `lat` are given as scalar, it is possible to provide a radius around
    location for which are extracted.

    radius :: radius around location given in rotated degrees
    data_path :: path to SMARTCARB wind fields
    method :: interpolation method (used by xr.DataArray.interp method)
    average :: average extracted winds

    Return xr.Dataset with wind components `U` and `V` as well as `wind speed`
    and `direction`.
    """
    lon = np.atleast_1d(lon).flatten()
    lat = np.atleast_1d(lat).flatten()

    if lon.size > 1 and radius is not None:
        raise ValueError(
            'Please provide lon and lat as scalar if radius is given.'
        )

    rlon, rlat, _ = np.transpose(
        DOMAIN.proj.transform_points(ccrs.PlateCarree(), lon, lat)
    )

    filename = time.round('H').strftime('SMARTCARB_winds_%Y%m%d%H.nc')
    filename = os.path.join(data_path, filename)

    with xr.open_dataset(filename) as nc:
        if radius is None:
            u = [float(nc['U_GNFR_A'].interp(rlon=rlon[i], rlat=rlat[i],
                                             method=method))
                 for i in range(rlon.size)]

            v = [float(nc['V_GNFR_A'].interp(rlon=rlon[i], rlat=rlat[i],
                                             method=method))
                 for i in range(rlon.size)]

        else:
            distance = np.sqrt((nc.rlon - rlon)**2 + (nc.rlat - rlat)**2)

            if np.any(distance <= radius):
                region = nc.where(distance <= radius, drop=True)

                lon = region.lon.values.flatten()
                lat = region.lat.values.flatten()
                u = region['U_GNFR_A'].values.flatten()
                v = region['V_GNFR_A'].values.flatten()
            else:
                lon = lon
                lat = lat
                u = np.array([np.nan])
                v = np.array([np.nan])

    wind = xr.Dataset()
    wind['lon'] = xr.DataArray(lon, dims='index')
    wind['lat'] = xr.DataArray(lat, dims='index')
    wind['U'] = xr.DataArray(u, dims='index')
    wind['V'] = xr.DataArray(v, dims='index')

    if average:
        u_std = wind['U'].std()
        v_std = wind['V'].std()
        wind = wind.mean('index')
        wind['U_std'] = u_std
        wind['V_std'] = v_std

    wind['speed'] = np.sqrt(wind.U**2 + wind.V**2)
    wind['speed_precision'] = 1.0 # FIXME
    wind['direction'] = ddeq.wind.calculate_wind_direction(wind.U, wind.V)

    return wind.where(np.isfinite(wind['U']), drop=True)


def read_era5(time, lon, lat, data_path='.',timesteps=1):

    lon = np.atleast_1d(lon).flatten()
    lat = np.atleast_1d(lat).flatten()

    u = np.zeros_like(lon)
    v = np.zeros_like(lon)

    for i in range(lon.size):
        era5 = get_era5_wind_profile(time, lon[i], lat[i],
                                     data_path=data_path,
                                     timesteps=timesteps)

        # TODO: use weights
        if i == 0:
            wind = era5.mean('level').expand_dims("index")
        else:
            wind = xr.concat([wind, era5.mean('level')], dim="index")

    wind = wind.rename({"longitude": "lon",
                        "latitude": "lat",
                        "u": "U",
                        "v": "V"})
    wind = wind.drop_vars(["t", "z"])
    wind = wind.reset_coords()
    
    wind['speed'] = np.sqrt(wind.U**2 + wind.V**2)
    wind['speed_precision'] = (wind.U.dims, np.full_like(wind.U, 1.0)) # FIXME
    wind['direction'] = ddeq.wind.calculate_wind_direction(wind.U, wind.V)

    return wind.where(np.isfinite(wind['U']), drop=True)


def read_coco2_era5(time, lon, lat, suffix=None, method='linear'):
    """
    Read ERA-5 data from CoCO2 project on ICOS-CP.
    """
    warnings.warn('CoCO2-specific function will be removed in future.',
                  DeprecationWarning, stacklevel=2)

    if suffix is None:
        filename = time.strftime(
            '/project/coco2/jupyter/WP4/ERA5/ERA5-gnfra-%Y%m%dt%H00.nc'
        )
    else:
        filename = time.strftime(
            f'/project/coco2/jupyter/WP4/ERA5/ERA5-gnfra-%Y%m%dt%H00_{suffix}.nc'
        )

    lon = np.atleast_1d(lon)
    lat = np.atleast_1d(lat)

    with xr.open_dataset(filename) as nc:

        u = [float(nc['U_GNFR_A'].interp(longitude=lon[i], latitude=lat[i],
                                         method=method))
             for i in range(lon.size)]

        v = [float(nc['V_GNFR_A'].interp(longitude=lon[i], latitude=lat[i],
                                         method=method))
                 for i in range(lon.size)]

    wind = xr.Dataset()
    wind['lon'] = xr.DataArray(lon, dims='index')
    wind['lat'] = xr.DataArray(lat, dims='index')
    wind['U'] = xr.DataArray(u, dims='index')
    wind['V'] = xr.DataArray(v, dims='index')

    wind['speed'] = np.sqrt(wind.U**2 + wind.V**2)
    wind['speed_precision'] = 1.0 # FIXME
    wind['direction'] = ddeq.wind.calculate_wind_direction(wind.U, wind.V)

    return wind


def calculate_wind_direction(u, v):
    return (270.0 - np.rad2deg(np.arctan2(v,u))) % 360


def download_era5_winds(time, data_path='.', overwrite=False,
                        timesteps=1):
    """
    Download ERA-5 data of a given time step.
    
    timesteps :: number of hours to download on the given day.
        If 24, a full day will be returned.
    """

    if timesteps == 24:
        era5_filename = os.path.join(data_path,
                                     time.strftime('ERA5-wind-%Y%m%d.nc'))
    else:
        era5_filename = os.path.join(data_path,
                                     time.strftime('ERA5-wind-%Y%m%d_%H00.nc'))

    if os.path.exists(era5_filename) and not overwrite:
        return era5_filename

    cds = cdsapi.Client()

    query = {
        'product_type': 'reanalysis',
        'format': 'netcdf',
        'variable': [
            'geopotential', 'temperature', 'u_component_of_wind',
            'v_component_of_wind',
        ],
        'pressure_level': [
            '700', '750', '775',
            '800', '825', '850',
            '875', '900', '925',
            '950', '975', '1000',
        ],
        "area": [90, -180, -90, 180], # north, east, south, west
    }

    query["date"] = time.strftime('%Y-%m-%d')
    
    if timesteps == 1:
        query["time"] = time.round('H').strftime('%H:00')
    elif timesteps == 24:
        query['time'] = [f"{hour:02d}:00" for hour in range(0, timesteps, 1)]
    else:
        query['time'] = [f"{hour:02d}:00" for hour in range(
            max(0,time.round('H').hour+1-timesteps),
            time.round('H').hour+1,
            1)]
        
    r = cds.retrieve('reanalysis-era5-pressure-levels', query, era5_filename)
    
    return era5_filename


def get_era5_wind_profile(time, lon, lat, data_path='.',
                          timesteps=1):
    
    era5_filename = download_era5_winds(time, data_path,
                                        overwrite=False, timesteps=timesteps)

    with xr.open_dataset(era5_filename) as era5:
        era5 = era5.load().sel(longitude=lon, latitude=lat,
                               method='nearest')

    return era5



def calculate_pressure_levels(surface_pressure, nlevels=37, which='middle'):

    filename= os.path.join(os.path.dirname(__file__), 'data',
                           'ERA5 L137 model level definitions.csv')

    level_definitions = pd.read_csv(filename, index_col=0)

    a = np.array(level_definitions['a [Pa]'])
    b = np.array(level_definitions['b'])

    # pressure at boundary of model layers
    ph = a + b * float(surface_pressure)

    # TODO: how to interpolate?
    #ph = np.interp(np.linspace(0,138,nlevels+1), np.arange(138), ph)

    # pressure at middle of model layers
    pf = 0.5 * (ph[1:] + ph[:-1])

    if which == 'middle':
        return pf
    elif which == 'boundary':
        return ph
    else:
        raise ValueError


def calculate_level_heights(temperature, pressure, which='middle'):
    """
    temperature at middle of layers
    pressure at boundary of layers
    """
    g = 9.80665
    R = 287.06

    p1 = pressure[1:]
    p2 = pressure[:-1]

    dh = R * temperature / g * np.log(p1 / p2)

    hh = np.concatenate([[0.0], dh[::-1].cumsum()])
    hh = hh[::-1]

    hf = 0.5 * (hh[1:] + hh[:-1])

    if which == 'middle':
        return hf
    elif which == 'boundary':
        return hh
    else:
        raise ValueError


def read(time, sources, product='ERA5', data_path='.', radius=0.05,
         timesteps=1):
    """\
    Read wind dataset for sources.

    Product: ERA5, SMARTCARB, ...
    """
    if isinstance(time, xr.DataArray):
        time = pd.Timestamp(time.to_pandas())
    elif isinstance(time, np.datetime64):
        time = pd.Timestamp(time)

    winds = {}
    for name, source in sources.groupby('source'):
        if product in ['SC', 'SMARTCARB']:
            winds[name] = read_smartcarb(
                time, source.lon_o.values, source.lat_o.values,
                data_path=data_path,
                radius=radius, method='linear', average=True
            )
        elif product in ['ERA-5', 'ERA5']:
            wind = read_era5(
                time, source.lon_o.values, source.lat_o.values,
                data_path=data_path, timesteps=timesteps
            )
            winds[name] = wind.rename_dims({'index': 'source'})
        else:
            raise ValueError

    winds = xr.concat(winds.values(), dim='source')
    winds['source'] = xr.DataArray(sources.source, dims='source')

    return winds



def read_wind_field(wind_data, source, lon, lat, radius):
    """
    Read wind at a certain point from TROPOMI or MicroHH files.
    
    Parameters:
    ----------
    wind_data: netCDF
        A file containing the wind information
    source: str
        The name of the source for which the wind is read in
    lon: float
        Longitude of the source
    lat: float
        Latitude of the source
    radius: int
        Radius of pixels around the source.
        
    Returns:
    wind_src: netCDF
        A file with the mean wind information of the kernel with radius 'radius'
    """
    
    if "UM" in wind_data.data_vars:
        wind_data = wind_data.rename({
            "UM": "U",
            "VM": "V"
        })
    elif "longitude" in wind_data.data_vars:
        wind_data = wind_data.rename({
            "longitude": "lon",
            "latitude": "lat"
        })

    wind_data['speed'] = np.sqrt(wind_data.U**2 + wind_data.V**2)
    wind_data['direction'] = ddeq.wind.calculate_wind_direction(wind_data.U, wind_data.V)
    
    lon_closest, lat_closest, dist = ddeq.misc.find_closest(wind_data, ("lon", "lat"), (lon, lat)) # get values for closest coordinate
    
    wind_src = xr.Dataset(
        coords = {
            "source": (["source"], np.array([source], dtype=object))
        },
        data_vars = {
            "lon": (
                ["source"],
                [wind_data.lon[lon_closest, lat_closest].values]
            ),
            "lat": (
                ["source"],
                [wind_data.lat[lon_closest, lat_closest].values]
            ),
            "U": (
                ["source"],
                [wind_data.U[lon_closest-radius:lon_closest+radius+1,
                                lat_closest-radius:lat_closest+radius+1].median().values],
                {"units": "m s-1"}
            ),
            "V": (
                ["source"],
                [wind_data.V[lon_closest-radius:lon_closest+radius+1,
                                lat_closest-radius:lat_closest+radius+1].median().values],
                {"units": "m s-1"}
            ),
            "speed": (
                ["source"],
                [wind_data.speed[lon_closest-radius:lon_closest+radius+1,
                                   lat_closest-radius:lat_closest+radius+1].median().values],
                {"units": "m s-1"}),
            "speed_precision": (
                ["source"],
                [wind_data.speed[lon_closest-radius:lon_closest+radius+1,
                                   lat_closest-radius:lat_closest+radius+1].std().values],
                {"units": "m s-1"}),
            "direction": (
                ["source"],
                [scipy.stats.circmean(wind_data.direction[lon_closest-radius:lon_closest+radius+1,
                                                           lat_closest-radius:lat_closest+radius+1].values.flatten(),
                                     high = 360, nan_policy="omit")],
                {"units": "°"}),
            "direction_precision": (
                ["source"],
                [scipy.stats.circstd(wind_data.direction[lon_closest-radius:lon_closest+radius+1,
                                                           lat_closest-radius:lat_closest+radius+1].values.flatten(),
                                     high = 360, nan_policy="omit")],
                {"units": "°"}),
        },
        attrs=dict(radius=f"{radius} pixels around determined pixel")
    )

    return wind_src


def integrate_wind(data, curves, winds, time, crs, pixel_size):
    """
    Integrate wind backwards in time
    
    """
    
    if isinstance(time, xr.DataArray):
        time = pd.Timestamp(time.to_pandas())
    elif isinstance(time, np.datetime64) or isinstance(time, str):
        time = pd.Timestamp(time)

    src = 0
    for source in data.source.values:
        
        if curves[source] is None:
            continue
            
        s = 0
        in_plume = True
        time_plume = time
        winds_int = winds.sel(time=time, method='nearest').sel(source=source)
        
        if curves[source] is None:
            in_plume = False
            angle = np.nan
        else:
            angle = ddeq.misc.compute_angle_between_curve_and_wind(curves[source],
                                                                   winds_int.direction,
                                                                   crs=crs)

        winds_int["distance"] = ((), 0, {"units": "m"})
        winds_int["in_plume"] = ((), True)
        winds_int["angle_between_curve_and_wind"] = ((), float(angle), {"units": "°"})
        w = float(winds_int.speed)

        while in_plume and pd.Timestamp(min(winds.time.values)):

            s += w * 60**2 # compute distance the air parcel has travelled in 1h
            nrow, nob, dist = ddeq.misc.find_closest(data.sel(source=source), ("xp","yp"), (s,0))

            # test if the parcel is still inside the plume. Use this as last value for interpolation
            if dist > np.sqrt(pixel_size**2 + pixel_size**2):
                in_plume = False

            time_plume = time_plume - pd.Timedelta(hours=1)
            
            # get wind data at the source of the plume
            winds_plume = winds.sel(time=time_plume, method='nearest').sel(source=source)
            winds_plume["lon"] = data.isel(nrows=nrow, nobs=nob).lon.values
            winds_plume["lat"] = data.isel(nrows=nrow, nobs=nob).lat.values
            winds_plume["distance"] = ((), s, {"units": "m"})
            winds_plume["in_plume"] = ((), in_plume)
            angle = ddeq.misc.compute_angle_between_curve_and_wind(curves[source],
                                                                   winds_plume.direction,
                                                                   crs=crs)
            winds_plume["angle_between_curve_and_wind"] = ((), float(angle), {"units": "°"})
            w = float(winds_plume.speed)

            if w == 0.0 or np.isnan(w):
                in_plume = False

            winds_int = xr.concat([winds_int, winds_plume], dim="time")
        
        if src == 0:
            winds_int_all = winds_int
            src += 1
        else:
            winds_int_all = xr.concat([winds_int_all, winds_int], dim="source")
    
    return winds_int_all


def interpolate_wind(data, winds, delta):
    """
    Interpolate wind to the same along plume distances as in 'data'.
    """
    
    src = 0
    for source in data.source.values:
        
        ## REMOVE LATER
        if source not in ["Belchatow", "Janschwalde", "Lipetsk", "Matimba"]:
            continue
        
        if "xp" not in data.sel(source=source).data_vars:
            return winds
        
        dmax = np.nanmax(data.sel(source=source).xp.values[data.sel(source=source).detected_plume]) + delta/2
        new_along = np.insert(np.arange(delta/2, int(dmax), delta), 0, -7000)

        winds_interp = xr.Dataset(
            coords = {
                "source": (["source"], np.array([source], dtype=object)),
                "along": (["along"], np.array(new_along, dtype=int), {"units": "m"})
            })

        if not len(winds.time) == 0:
            for varname, da in winds.data_vars.items():
                if varname != "distance":
                    new_var = np.interp(
                        np.insert(np.arange(delta/2, int(dmax), delta), 0, 0),
                        winds.distance.values.flatten(),
                        da.values.flatten())
                    winds_interp[varname] = ("along", new_var, {"units": da.attrs.get('units', "")})
            
            int_time = np.interp(
                np.insert(np.arange(delta/2, int(dmax), delta), 0, 0),
                winds.distance.values.flatten(),
                abs((winds.time.values.flatten() - 
                     np.nanmax(winds.time.values)) / 1e9).astype(int))
            winds_interp["int_time"] = ("along", int_time, {"units": "s"})
                    
    if src == 0:
        winds_interp_all = winds_interp
        src += 1
    else:
        winds_interp_all = xr.concat([winds_interp_all, winds_interp], dim="source")
            
    return winds_interp_all

